#!/usr/bin/python3
"""
Module fritzToInfluxHA

This module reads data from Fritz!Box Home Automation modules
and and stores related measurement date in an InfluxDB
"""
import argparse
import hashlib

parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=
"""
This program calculates the response value for Fritz!Box login
from challenge and password
"""
)
parser.add_argument("-c", "--challenge", help="Challenge")
parser.add_argument("-p", "--password", help="Password")

args = parser.parse_args()

challenge = None
password = None

if args.challenge:
    challenge = args.challenge
if args.password:
    password = args.password

if not challenge or not password:
    print("challenge or password missing")
else:
    md5 = hashlib.md5()
    md5.update(challenge.encode('utf-16le'))
    md5.update('-'.encode('utf-16le'))
    md5.update(password.encode('utf-16le'))
    response = challenge + '-' + md5.hexdigest()

    print("response: " + response)





#!/usr/bin/python3
#MIT License
#
#Copyright (c) 2022 signag
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
"""Module FritzBox

This module includes classes for an abstraction of a Fritz!Box.
"""
import requests
import hashlib
import os
import xml.etree.ElementTree as ET
from enum import Enum
import datetime
import influxdb_client
from .FritzHaDevice import FritzHaDevice
from .FritzHaDevice import FritzHaDeviceInfluxWriteError

#Setup logging
import logging
import logging_plus
logger = logging_plus.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Supported device types
class HaDeviceType(Enum):
    SWITCH = 1
    REPEATER = 2
    LAMP = 3
    UNKNOWN = 9

class FritzBoxError(Exception):
    """
    Base exception class for this module
    """
    def __init__(self):
        self.message = "Fritz!Box general error"

class FritzBoxIgnoreableError(FritzBoxError):
    """
    Base exception class for this module
    """
    def __init__(self):
        self.message = "Fritz!Box ignoreable error"

class FritzBoxConnectionError(FritzBoxIgnoreableError):
    """
    Fritzbox Connection exception class for this module
    """
    def __init__(self):
        self.message = "Fritz!Box cannot be reached"

class FritzBoxLoginError(FritzBoxError):
    """
    Fritzbox Login exception class for this module
    """
    def __init__(self):
        self.message = "Fritz!Box login failed"

class FritzBoxNoDeviceError(FritzBoxError):
    """
    Fritzbox exception class for this module
    """
    def __init__(self):
        self.message = "No devices found on Fritz!Box"

class FritzBox:
    """
    Class representing a Fritz!Box
    """
    def __init__(self, url, user, pwd):
        """
        Constructor for Fritz!Box
        """
        self.url = url
        if self.url[-1] != "/":
            self.url = self.url + "/"

        self.sid = "0000000000000000"
        self.user = user
        self.pwd = pwd
        self.devices = []

        self.loginSuccess = False

        # Login
        try:
            self.login()
            self.loginSuccess = True

            # Get list of devices
            self.getHaDevices()

            ### Test: handling if no devices were found
            ### Test Start
            #self.devices = []
            ### Test End
            if len(self.devices) == 0:
                logger.error("No devices found for Fritz!Box")
                raise FritzBoxNoDeviceError
                
        except FritzBoxError:
            raise

    def __del__(self):
        self.terminate()

    def terminate(self):
        self.logoff()

    def login(self):
        """
        Login with Session-ID
        """
        #Try login with current sid
        theUrl = self.url + "login_sid.lua" + "?sid=" + self.sid
        resp = self.sendRequest(theUrl)
        root = ET.fromstring(resp)
        if root.findtext("SID") == "0000000000000000":
            #invalid SID. Need to get new SID
            challenge = root.findtext("Challenge")
            if challenge:
                self.getSid(challenge)
            else:
                raise FritzBoxLoginError

    def logoff(self):
        """
        Log off from Fritz!Box
        """
        theUrl = self.url + "?logout=1&sid=" + self.sid
        self.sendRequest(theUrl)


    def getSid(self, challenge):
        """
        Get the session ID
        """
        md5 = hashlib.md5()
        md5.update(challenge.encode('utf-16le'))
        md5.update('-'.encode('utf-16le'))
        md5.update(self.pwd.encode('utf-16le'))
        response = challenge + '-' + md5.hexdigest()
        theUrl = f"{self.url}login_sid.lua?username={self.user}&response={response}"
        resp = self.sendRequest(theUrl)
        if resp:
            root = ET.fromstring(resp)
            self.sid = root.findtext("SID")
            logger.debug("SID: %s", self.sid)
        else:
            raise FritzBoxLoginError

    def sendRequest(self, url):
        """
        Send a request with given URL and return response
        """
        logger.debug("Request URL: %s", url)
        try:
            resp = requests.get(url)
            if resp.status_code == requests.codes.OK:
                respTxt = resp.text.strip()
                logger.debug("Response: %s", respTxt)
                return respTxt
            else:
                logger.error("HTTP request [%s] failed with status code %s reason %s", resp.url, resp.status_code, resp.reason)
                resp.raise_for_status
                return None
        except (requests.ConnectionError, \
                requests.ConnectTimeout, \
                requests.ReadTimeout \
        ):
            if self.loginSuccess:
                # Ignore connection error if FritzBox is temporarily not reacheable
                raise FritzBoxConnectionError
            else:
                raise

    def getHaDevices(self):
        """
        Get the Home Automation devices registered for the Fritz!Box
        """
        # Get the device list infos
        # This is preferred to get switch list because it includes all devices
        theUrl = self.url + "webservices/homeautoswitch.lua" + "?switchcmd=getdevicelistinfos&sid=" + self.sid
        resp = self.sendRequest(theUrl)
        root = ET.fromstring(resp)

        # Loop through devices
        for dev in root:
            ain = dev.attrib['identifier']
            ain = ain.strip()
            ain = ain.replace(" ", "")
            product = dev.attrib['productname']
            haDev = FritzHaDevice(ain)

            if product == "FRITZ!DECT 200" or product == "FRITZ!DECT 210":
                haDev.type = HaDeviceType.SWITCH
                haDev.hasState = True
                switch = dev.find("switch")
                haDev.state = switch.findtext("state")
            elif product == "FRITZ!DECT Repeater 100":
                haDev.type = HaDeviceType.REPEATER
                haDev.hasTemperature = True
            else:
                haDev.type = HaDeviceType.UNKNOWN

            haDev.name = dev.findtext("name")
            haDev.present = dev.findtext("present")

            if dev.find("powermeter"):
                haDev.hasPower = True
            if dev.find("temperature"):
                haDev.hasTemperature = True

            self.devices.append(haDev)

    def completeDeviceData(self, data):
        """
        Complete device data with given data        
        """
        for dev in self.devices:
            ain = dev.ain
            for ref in data:
                if ref["ain"] == ain:
                    dev.completeData(ref)
    
    def evaluateDeviceInfo(self):
        """
        Query device info from Fritzbox and update devices with measurements
        """
        # Reset device upToDate status
        for sdev in self.devices:
            sdev.upToDate = False
            
        try:
            measurementTime = datetime.datetime.now()
            theUrl = self.url + "webservices/homeautoswitch.lua" + "?switchcmd=getdevicelistinfos&sid=" + self.sid
            resp = self.sendRequest(theUrl)
            if not resp:
                # In case of request error: login with new SID
                self.login()
                theUrl = self.url + "webservices/homeautoswitch.lua" + "?switchcmd=getdevicelistinfos&sid=" + self.sid
                resp = self.sendRequest(theUrl)
                if not resp:
                    # In case of repeated error throw exception
                    logger.error("Error sending request for getdevicelistinfos after successful login")
                    raise FritzBoxError

            root = ET.fromstring(resp)
            for dev in root:
                ain = dev.attrib['identifier']
                ain = ain.strip()
                ain = ain.replace(" ", "")
                device = None
                for sdev in self.devices:
                    if sdev.ain == ain:
                        device = sdev
                        break
                if device:
                    powermeter = dev.find("powermeter")
                    if powermeter:
                        voltage = powermeter.findtext("voltage")
                        if voltage:
                            device.voltage = int(voltage)/1000
                        power = powermeter.findtext("power")
                        if power:
                            device.power = int(power)/1000
                        energy = powermeter.findtext("energy")
                        if energy:
                            device.energy = int(energy)/1000
                    temperature = dev.find("temperature")
                    if temperature:
                        celsius = temperature.findtext("celsius")
                        if celsius:
                            device.temperature = int(celsius)/10
                    device.measurementTime = measurementTime
                    device.upToDate = True

        except FritzBoxError as error:
            raise

    def writeDataToCsv(self, fp):
        """
        Write measurement values to a csv file
        """
        f = None
        newFile=True
        if os.path.exists(fp):
            newFile = False
        if newFile:
            f = open(fp, 'w')
        else:
            f = open(fp, 'a')
        logger.debug("File opened: %s", fp)

        sep = ","
        if newFile:
            txt = "Time" + sep \
                + "AIn" + sep \
                + "Type" + sep \
                + "Name" + sep \
                + "Location" + sep \
                + "Sublocation" + sep \
                + "State" + sep \
                + "Present" + sep \
                + "Voltage" + sep \
                + "Power" + sep \
                + "Energy" + sep \
                + "Temperature" \
                + "\n"
            f.write(txt)

        for dev in self.devices:
            if dev.upToDate:
                txt = ""
                if dev.measurementTime:
                    ts = dev.measurementTime.strftime("%Y-%m-%d %H:%M:%S.%f")
                    txt = txt + ts + sep
                else:
                    txt = txt + sep
                if dev.ain:
                    txt = txt + dev.ain + sep
                else:
                    txt = txt + sep
                if dev.type:
                    txt = txt + dev.type.name + sep
                else:
                    txt = txt + sep
                if dev.name:
                    txt = txt + dev.name + sep
                else:
                    txt = txt + sep
                if dev.location:
                    txt = txt + dev.location + sep
                else:
                    txt = txt + sep
                if dev.sublocation:
                    txt = txt + dev.sublocation + sep
                else:
                    txt = txt + sep
                if dev.state:
                    txt = txt + dev.state + sep
                else:
                    txt = txt + sep
                if dev.present:
                    txt = txt + dev.present + sep
                else:
                    txt = txt + sep
                if dev.voltage:
                    txt = txt + format(dev.voltage) + sep
                else:
                    txt = txt + sep
                if dev.power:
                    txt = txt + format(dev.power) + sep
                else:
                    txt = txt + sep
                if dev.energy:
                    txt = txt + format(dev.energy) + sep
                else:
                    txt = txt + sep
                if dev.temperature:
                    txt = txt + format(dev.temperature)
                txt = txt + "\n"
                f.write(txt)

        f.close()

    def writeDataToInflux(self, write_api, org, bucket):
        """
        Write measurements to InfluxDB
        """
        try:
            for dev in self.devices:
                if dev.isMonitored:
                    dev.writeMeasurmentsToInfluxDB(write_api, org, bucket)
        except FritzHaDeviceInfluxWriteError as error:
            logger.error("Error: %s", error.message)
            raise FritzBoxIgnoreableError
            


#!/usr/bin/python3
"""Module FritzHaDevice

This module includes classes for an abstraction of a Fritz Home Automation device.
"""

#Setup logging
from math import fabs
import influxdb_client
from influxdb_client.client.write_api import WritePrecision
import logging
import logging_plus

logger = logging_plus.getLogger(__name__)
logger.addHandler(logging.NullHandler())

class FritzHaDeviceError(Exception):
    """
    Base exception class for this module
    """
    def __init__(self):
        self.message = "FritzHaDeviceError"

class FritzHaDeviceInfluxWriteError(FritzHaDeviceError):
    """
    Base exception class for this module
    """
    def __init__(self):
        self.message = "Error while writing data to InfluxDB"

class FritzHaDevice:
    """
    Class representing a Fritz Home Automation device
    """
    def __init__(self, ain):
        """
        Constructor for Fritz device
        """
        self.ain = ain
        self.type = None
        self.name = None
        self.location = None
        self.sublocation = None
        self.state = None
        self.present = None

        self.upToDate = False
        self.voltage = None
        self.power = None
        self.energy = None
        self.temperature = None
        self.measurementTime = None

        self.hasState = False
        self.hasTemperature = False
        self.hasPower = False

        self.measureVoltage = False
        self.measurements = {}

        self.isMonitored = False

    def __del__(self):
        pass

    def completeData(self, data):
        """
        Complete data with given data
        """
        if "location" in data:
            self.location = data["location"]
        if "sublocation" in data:
            self.sublocation = data["sublocation"]
        if "measurements" in data:
            self.measurements = data["measurements"]
        self.isMonitored = True            

    def writeMeasurmentsToInfluxDB(self, write_api, org, bucket):
        ts = self.measurementTime.strftime("%Y-%m-%dT%H:%M:%S.%f+02")
        try:
            if self.upToDate:
                if "voltage" in self.measurements:
                    if self.measurements["voltage"] and self.voltage:
                        point = influxdb_client.Point("voltage") \
                            .tag("ain", self.ain) \
                            .tag("location", self.location) \
                            .tag("sublocation", self.sublocation) \
                            .tag("state", self.state) \
                            .field("value", self.voltage)
                        write_api.write(bucket=bucket, org=org, record=point)

                if "power" in self.measurements:
                    if self.measurements["power"] and self.power:
                        point = influxdb_client.Point("power") \
                            .tag("ain", self.ain) \
                            .tag("location", self.location) \
                            .tag("sublocation", self.sublocation) \
                            .tag("state", self.state) \
                            .field("value", self.power)
                        write_api.write(bucket=bucket, org=org, record=point)

                if "energy" in self.measurements:
                    if self.measurements["energy"] and self.energy:
                        point = influxdb_client.Point("energy") \
                            .tag("ain", self.ain) \
                            .tag("location", self.location) \
                            .tag("sublocation", self.sublocation) \
                            .tag("state", self.state) \
                            .field("value", self.energy)
                        write_api.write(bucket=bucket, org=org, record=point)

                if "temperature" in self.measurements:
                    if self.measurements["temperature"] and self.temperature:
                        state = self.state
                        if not state:
                            state = "1"
                        point = influxdb_client.Point("temperature") \
                            .tag("ain", self.ain) \
                            .tag("location", self.location) \
                            .tag("sublocation", self.sublocation) \
                            .tag("state", state) \
                            .field("value", self.temperature)
                        write_api.write(bucket=bucket, org=org, record=point)
        except Exception:
            raise FritzHaDeviceInfluxWriteError



#!/usr/bin/python3
"""
Module fritzToInfluxHA

This module periodically reads data from Fritz!Box Home Automation modules
and and stores related measurement date in an InfluxDB
"""

import time
import datetime
import math
import os.path
import json
import influxdb_client
from influxdb_client.client.write_api import SYNCHRONOUS
from fritz.FritzBox import FritzBox, FritzBoxError, FritzBoxIgnoreableError

# Set up logging
import logging
from logging.config import dictConfig
import logging_plus
logger = logging_plus.getLogger("main")

testRun = False
servRun = False

# Configuration defaults
cfgFile = ""
cfg = {
    "measurementInterval": 120,
    "FritzBoxURL" : "http://fritz.box/",
    "FritzBoxUser" : None,
    "FritzBoxPassword" : None,
    "InfluxOutput" : False,
    "InfluxURL" : None,
    "InfluxOrg" : None,
    "InfluxToken" : None,
    "InfluxBucket" : None,
    "csvOutput" : False,
    "csvFile" : "",
    "devices" : []
}

# Constants
CFGFILENAME = "fritzToInfluxHA.json"

def getCl():
    """
    getCL: Get and process command line parameters
    """

    import argparse
    import os.path

    global logger
    global testRun
    global servRun
    global cfgFile

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=
    """
    This program periodically reads data from Fritz!Box HA components
    and stores these as measurements in an InfluxDB database.

    If not otherwises specified on the command line, a configuration file
       fritzToInfluxHA.json
    will be searched sequentially under ./tests/data, $HOME/.config or /etc.

    This configuration file specifies credentials for Fritz!Box access,
    the devices to read from, the connection to the InfluxDB and other runtime parameters.
    """
    )
    parser.add_argument("-t", "--test", action = "store_true", help="Test run - single cycle - no wait")
    parser.add_argument("-s", "--service", action = "store_true", help="Run as service - special logging")
    parser.add_argument("-l", "--log", action = "store_true", help="Shallow (module) logging")
    parser.add_argument("-L", "--Log", action = "store_true", help="Deep logging")
    parser.add_argument("-F", "--Full", action = "store_true", help="Full logging")
    parser.add_argument("-f", "--file", help="Logging configuration from specified JSON dictionary file")
    parser.add_argument("-v", "--verbose", action = "store_true", help="Verbose - log INFO level")
    parser.add_argument("-c", "--config", help="Path to config file to be used")

    args = parser.parse_args()

    # Disable logging
    logger = logging_plus.getLogger("main")
    logger.addHandler(logging.NullHandler())
    fLogger = logging_plus.getLogger(FritzBox.__module__)
    fLogger.addHandler(logging.NullHandler())
    rLogger = logging_plus.getLogger()
    rLogger.addHandler(logging.NullHandler())

    # Set handler and formatter to be used
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s')
    formatter2 = logging.Formatter('%(asctime)s %(name)-33s %(levelname)-8s %(message)s')
    handler.setFormatter(formatter)

    if args.log:
        # Shallow logging
        handler.setFormatter(formatter2)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

    if args.Log:
        # Deep logging
        handler.setFormatter(formatter2)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        fLogger.addHandler(handler)
        fLogger.setLevel(logging.DEBUG)

    if args.Full:
        # Full logging
        handler.setFormatter(formatter2)
        rLogger.addHandler(handler)
        rLogger.setLevel(logging.DEBUG)
        # Activate logging of function entry and exit
        logging_plus.registerAutoLogEntryExit()

    if args.file:
        # Logging configuration from file
        logDictFile = args.file
        if not os.path.exists(logDictFile):
            raise ValueError("Logging dictionary file from command line does not exist: " + logDictFile)

        # Load dictionary
        with open(logDictFile, 'r') as f:
            logDict = json.load(f)

        # Set config file for logging
        dictConfig(logDict)
        logger = logging.getLogger()
        # Activate logging of function entry and exit
        #logging_plus.registerAutoLogEntryExit()

    # Explicitly log entry
    if args.Log or args.Full:
        logger.logEntry("getCL")
    if args.log:
        logger.debug("Shallow logging (main only)")
    if args.Log:
        logger.debug("Deep logging")
    if args.file:
        logger.debug("Logging dictionary from %s", logDictFile)

    if args.verbose or args.service:
        if not args.log and not args.Log and not args.Full:
            logger.addHandler(handler)
            logger.setLevel(logging.INFO)
            fLogger.addHandler(handler)
            fLogger.setLevel(logging.WARNING)

    if args.test:
        testRun = True

    if args.service:
        servRun = True

    if testRun:    
        logger.debug("Test run mode activated")
    else:
        logger.debug("Test run mode deactivated")
        
    if servRun:    
        logger.debug("Service run mode activated")
    else:
        logger.debug("Service run mode deactivated")

    if args.config:
        cfgFile = args.config
        logger.debug("Config file: %s", cfgFile)
    else:
        logger.debug("No Config file specified on command line")

    if args.Log or args.Full:
        logger.logExit("getCL")

def getConfig():
    """
    Get configuration for fritzToInfluxHA
    """
    global cfgFile
    global cfg
    global logger

    # Check config file from command line
    if cfgFile != "":
        if not os.path.exists(cfgFile):
            raise ValueError("Configuration file from command line does not exist: ", cfgFile)
        logger.info("Using cfgFile from command line: %s", cfgFile)

    if cfgFile == "":
        # Check for config file in ./tests/data directory
        curDir = os.path.dirname(os.path.realpath(__file__))
        curDir = os.path.dirname(curDir)
        cfgFile = curDir + "/tests/data/" + CFGFILENAME
        if not os.path.exists(cfgFile):
            # Check for config file in /etc directory
            logger.info("Config file not found: %s", cfgFile)
            cfgFile = ""

    if cfgFile == "":
        # Check for config file in home directory
        homeDir = os.environ['HOME']
        cfgFile = homeDir + "/.config/" + CFGFILENAME
        if not os.path.exists(cfgFile):
            # Check for config file in /etc directory
            logger.info("Config file not found: %s", cfgFile)
            cfgFile = "/etc/" + CFGFILENAME
            if not os.path.exists(cfgFile):
                logger.info("Config file not found: %s", cfgFile)
                cfgFile = ""

    if cfgFile == "":
        # No cfg available 
        logger.info("No config file available. Using default configuration")
    else:
        logger.info("Using cfgFile: %s", cfgFile)
        with open(cfgFile, 'r') as f:
            conf = json.load(f)
            if "measurementInterval" in conf:
                cfg["measurementInterval"] = conf["measurementInterval"]
            if "FritzBoxURL" in conf:
                cfg["FritzBoxURL"] = conf["FritzBoxURL"]
            if "FritzBoxUser" in conf:
                cfg["FritzBoxUser"] = conf["FritzBoxUser"]
            if "FritzBoxPassword" in conf:
                cfg["FritzBoxPassword"] = conf["FritzBoxPassword"]
            if "InfluxOutput" in conf:
                cfg["InfluxOutput"] = conf["InfluxOutput"]
            if "InfluxURL" in conf:
                cfg["InfluxURL"] = conf["InfluxURL"]
            if "InfluxOrg" in conf:
                cfg["InfluxOrg"] = conf["InfluxOrg"]
            if "InfluxToken" in conf:
                cfg["InfluxToken"] = conf["InfluxToken"]
            if "InfluxBucket" in conf:
                cfg["InfluxBucket"] = conf["InfluxBucket"]
            if "csvOutput" in conf:
                cfg["csvOutput"] = conf["csvOutput"]
            if "csvFile" in conf:
                cfg["csvFile"] = conf["csvFile"]
            if cfg["csvFile"] == "":
                cfg["csvOutput"] = False
            if "devices" in conf:
                cfg["devices"] = conf["devices"]

    logger.info("Configuration:")
    logger.info("    measurementInterval:%s", cfg["measurementInterval"])
    logger.info("    FritzBoxURL:%s", cfg["FritzBoxURL"])
    logger.info("    FritzBoxUser:%s", cfg["FritzBoxUser"])
    logger.info("    FritzBoxPassword:%s", cfg["FritzBoxPassword"])
    logger.info("    InfluxOutput:%s", cfg["InfluxOutput"])
    logger.info("    InfluxURL:%s", cfg["InfluxURL"])
    logger.info("    InfluxOrg:%s", cfg["InfluxOrg"])
    logger.info("    InfluxToken:%s", cfg["InfluxToken"])
    logger.info("    InfluxBucket:%s", cfg["InfluxBucket"])
    logger.info("    csvOutput:%s", cfg["csvOutput"])
    logger.info("    csvFile:%s", cfg["csvFile"])
    logger.info("    Devices:%s", len(cfg["devices"]))
    for ind in range(0, len(cfg["devices"])):
        dev = cfg["devices"][ind]
        logger.info("       %s (%s - %s)", dev["ain"], dev["location"], dev["sublocation"])


def waitForNextCycle():
    """
    Wait for next measurement cycle.

    This function assures that measurements are done at specific times depending on the specified interval
    In case that measurementInterval is an integer multiple of 60, the waiting time is calculated in a way,
    that one measurement is done every full hour.
    """
    global cfg

    if (cfg["measurementInterval"] % 60 == 0)\
    or (cfg["measurementInterval"] % 120 == 0)\
    or (cfg["measurementInterval"] % 240 == 0)\
    or (cfg["measurementInterval"] % 300 == 0)\
    or (cfg["measurementInterval"] % 360 == 0)\
    or (cfg["measurementInterval"] % 600 == 0)\
    or (cfg["measurementInterval"] % 720 == 0)\
    or (cfg["measurementInterval"] % 900 == 0)\
    or (cfg["measurementInterval"] % 1200 == 0)\
    or (cfg["measurementInterval"] % 1800 == 0):
        tNow = datetime.datetime.now()
        seconds = 60 * tNow.minute
        period = math.floor(seconds/cfg["measurementInterval"])
        waitTimeSec = (period + 1) * cfg["measurementInterval"] - (60 * tNow.minute + tNow.second + tNow.microsecond / 1000000)
        logger.debug("At %s waiting for %s sec.", datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S,"), waitTimeSec)
        time.sleep(waitTimeSec)
    elif (cfg["measurementInterval"] % 2 == 0)\
      or (cfg["measurementInterval"] % 4 == 0)\
      or (cfg["measurementInterval"] % 5 == 0)\
      or (cfg["measurementInterval"] % 6 == 0)\
      or (cfg["measurementInterval"] % 10 == 0)\
      or (cfg["measurementInterval"] % 12 == 0)\
      or (cfg["measurementInterval"] % 15 == 0)\
      or (cfg["measurementInterval"] % 20 == 0)\
      or (cfg["measurementInterval"] % 30 == 0):
            tNow = datetime.datetime.now()
            seconds = 60 * tNow.minute + tNow.second
            period = math.floor(seconds/cfg["measurementInterval"])
            waitTimeSec = (period + 1) * cfg["measurementInterval"] - seconds
            logger.debug("At %s waiting for %s sec.", datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S,"), waitTimeSec)
            time.sleep(waitTimeSec)
    else:
        waitTimeSec =cfg["measurementInterval"]
        logger.debug("At %s waiting for %s sec.", datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S,"), waitTimeSec)
        time.sleep(waitTimeSec)

def logDeviceInconsistencies(cfgDefs, fritzDevs):
    for dev in fritzDevs:
        if not dev.isMonitored:
            logger.error("Missing configuretion for device ain=%s name=%s", dev.ain, dev.name)
    for ind in range(0, len(cfg["devices"])):
        devc = cfg["devices"][ind]
        cnt = 0
        for dev in fritzDevs:
            if dev.ain == devc["ain"]:
                cnt = cnt + 1
                break
        if cnt == 0:
            logger.error("No device found for configuration ain=%s", devc["ain"])

#============================================================================================
# Start __main__
#============================================================================================
#
# Get Command line options
getCl()

logger.info("=============================================================")
logger.info("fritzToInfluxHA started")
logger.info("=============================================================")

# Get configuration
getConfig()

fb = None
influxClient = None
influxWriteAPI = None

try:
    # Log in to FritzBox
    fb = FritzBox(cfg["FritzBoxURL"], cfg["FritzBoxUser"], cfg["FritzBoxPassword"])
    logger.debug("FritzBox fb instantiated")

    # Complete device data from configiration data
    fb.completeDeviceData(cfg["devices"])
    logger.debug("Device data completed from config for %s devices", len(cfg["devices"]))

    # Log inconsistencies between configured devices and devices found on FritzBox
    logDeviceInconsistencies(cfg["devices"], fb.devices)

    # Instatntiate InfluxDB access
    if cfg["InfluxOutput"]:
        influxClient = influxdb_client.InfluxDBClient(
            url=cfg["InfluxURL"],
            token=cfg["InfluxToken"],
            org=cfg["InfluxOrg"]
        )
        influxWriteAPI = influxClient.write_api(write_options=SYNCHRONOUS)
        logger.debug("Influx interface instantiated")

    noWait = False
    stop = False

except FritzBoxError as error:
    logger.critical("Unexpected error: %s", error.message)
    stop = True
    fb = None
    influxClient = None
    influxWriteAPI = None

failcount = 0
while not stop:
    try:
        # Wait unless noWait is set in case of sensor error.
        # Akip waiting for test run
        if not noWait and not testRun:
            waitForNextCycle()
        noWait = False

        ### Test FritzBox down (simulated through invalid URL)
        ### Start Test
        #if failcount == 0:
        #    testRun = False
        #    # Make URL invalid
        #    fb.url = "http://fritzy.box"
        #    # Make sid invalid which would be the case for a FritzBox update
        #    fb.sid = "xyz"
        #if failcount == 2:
        #    fb.url = cfg["FritzBoxURL"]
        #    testRun = True
        ### End Test

        # Get measurements for all devices
        fb.evaluateDeviceInfo()
        if not servRun:
            logger.info("Measurement completed")

        # Write data to CSV
        if cfg["csvOutput"]:
            fp = cfg["csvFile"]
            fb.writeDataToCsv(fp)

        # Write data to InfluxDB
        if cfg["InfluxOutput"]:
            fb.writeDataToInflux(influxWriteAPI, cfg["InfluxOrg"], cfg["InfluxBucket"])
            if not servRun:
                logger.info("Data written to InfluxDB")

        if testRun:
            # Stop in case of test run
            stop = True

    except FritzBoxIgnoreableError as error:
        failcount = failcount + 1
        logger.error("Ignored FritzBoxIgnoreableError (%s): %s", failcount, error.message)

        noWait = True
        if testRun:
            # Stop in case of test run
            stop = True
        else:
            time.sleep(2.0)
            continue

    except FritzBoxError as error:
        logger.critical("Unexpected error: %s", error.message)
        if fb:
            del fb
        if influxClient:
            del influxClient
        if influxWriteAPI:
            del influxWriteAPI

    except Exception as e:
        logger.critical("Unexpected error (%s): %s", e.__class__, e.__cause__)
        if fb:
            del fb
        if influxClient:
            del influxClient
        if influxWriteAPI:
            del influxWriteAPI
        raise error

    except KeyboardInterrupt:
        if fb:
            del fb
        if influxClient:
            del influxClient
        if influxWriteAPI:
            del influxWriteAPI

if fb:
    del fb
if influxClient:
    del influxClient
if influxWriteAPI:
    del influxWriteAPI

logger.info("=============================================================")
logger.info("fritzToInfluxHA terminated")
logger.info("=============================================================")
